#!/usr/bin/env python
"""
 *  Copyright Synerty Pty Ltd 2020
 *
 *  This sw_upload is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this sw_upload are reserved by
 *  Synerty Pty Ltd
 *
"""

import logging
import os

from peek_platform.util.LogUtil import setupPeekLogger, updatePeekLoggerHandlers, \
    setupLoggingToSysloyServer
from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_server import importPackages
from peek_server.storage import setupDbConn
from peek_server.storage.DeclarativeBase import metadata
from pytmpdir.Directory import DirSettings
from txhttputil.site.FileUploadRequest import FileUploadRequest
from vortex.DeferUtil import vortexLogFailure
from vortex.VortexFactory import VortexFactory

setupPeekLogger(peekServerName)

from twisted.internet import reactor, defer

from txhttputil.site.SiteUtil import setupSite

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
# Set the parallelism of the database and reactor

def setupPlatform():
    from peek_platform import PeekPlatformConfig
    PeekPlatformConfig.componentName = peekServerName

    # Tell the platform classes about our instance of the pluginSwInstallManager
    from peek_server.server.sw_install.PluginSwInstallManager import \
        PluginSwInstallManager
    PeekPlatformConfig.pluginSwInstallManager = PluginSwInstallManager()

    # Tell the platform classes about our instance of the PeekSwInstallManager
    from peek_server.server.sw_install.PeekSwInstallManager import PeekSwInstallManager
    PeekPlatformConfig.peekSwInstallManager = PeekSwInstallManager()

    # Tell the platform classes about our instance of the PeekLoaderBase
    from peek_server.plugin.ServerPluginLoader import ServerPluginLoader
    PeekPlatformConfig.pluginLoader = ServerPluginLoader()

    # The config depends on the componentName, order is important
    from peek_server.PeekServerConfig import PeekServerConfig
    PeekPlatformConfig.config = PeekServerConfig()

    # Initialise the recovery user
    # noinspection PyStatementEffect
    PeekPlatformConfig.config.adminUser
    # noinspection PyStatementEffect
    PeekPlatformConfig.config.adminPass

    # Update the version in the config file
    from peek_server import __version__
    PeekPlatformConfig.config.platformVersion = __version__

    # Set default logging level
    logging.root.setLevel(PeekPlatformConfig.config.loggingLevel)
    updatePeekLoggerHandlers(PeekPlatformConfig.componentName,
                             PeekPlatformConfig.config.loggingRotateSizeMb,
                             PeekPlatformConfig.config.loggingRotationsToKeep,
                             PeekPlatformConfig.config.logToStdout)

    if PeekPlatformConfig.config.loggingLogToSyslogHost:
        setupLoggingToSysloyServer(PeekPlatformConfig.config.loggingLogToSyslogHost,
                                   PeekPlatformConfig.config.loggingLogToSyslogPort,
                                   PeekPlatformConfig.config.loggingLogToSyslogFacility)

    # Enable deferred debugging if DEBUG is on.
    if logging.root.level == logging.DEBUG:
        defer.setDebugging(True)

    # If we need to enable memory debugging, turn that on.
    if PeekPlatformConfig.config.loggingDebugMemoryMask:
        from peek_platform.util.MemUtil import setupMemoryDebugging
        setupMemoryDebugging(PeekPlatformConfig.componentName,
                             PeekPlatformConfig.config.loggingDebugMemoryMask)

    # Set the reactor thread count
    reactor.suggestThreadPoolSize(PeekPlatformConfig.config.twistedThreadPoolSize)

    # Set paths for the Directory object
    DirSettings.defaultDirChmod = PeekPlatformConfig.config.DEFAULT_DIR_CHMOD
    DirSettings.tmpDirPath = PeekPlatformConfig.config.tmpPath
    FileUploadRequest.tmpFilePath = PeekPlatformConfig.config.tmpPath

    # Configure the celery app
    from peek_platform.ConfigCeleryApp import configureCeleryApp
    from peek_plugin_base.worker.CeleryApp import celeryApp
    configureCeleryApp(celeryApp, PeekPlatformConfig.config, forCaller=True)


def startListening():
    from peek_server.backend.AdminSiteResource import setupAdminSite, adminSiteRoot
    from peek_server.backend.auth.AdminAuthChecker import AdminAuthChecker
    from peek_server.backend.auth.AdminAuthRealm import AdminAuthRealm
    from peek_platform import PeekPlatformConfig

    setupAdminSite()

    adminAuthChecker = AdminAuthChecker()
    adminAuthRealm = AdminAuthRealm(adminSiteRoot, adminAuthChecker)

    adminSiteCfg = PeekPlatformConfig.config.adminHttpServer
    setupSite("Peek Admin",
              adminAuthRealm,
              portNum=adminSiteCfg.sitePort,
              enableLogin=False,
              redirectFromHttpPort=adminSiteCfg.redirectFromHttpPort,
              sslCertFilePath=adminSiteCfg.sslCertFilePath,
              sslKeyFilePath=adminSiteCfg.sslKeyFilePath)

    from peek_server.server.PlatformSiteResource import setupPlatformSite
    from peek_server.server.PlatformSiteResource import platformSiteRoot

    setupPlatformSite()

    platformCfg = PeekPlatformConfig.config.platformHttpServer
    setupSite("Peek Platform Data Exchange",
              platformSiteRoot,
              portNum=platformCfg.sitePort,
              enableLogin=False,
              redirectFromHttpPort=platformCfg.redirectFromHttpPort,
              sslCertFilePath=platformCfg.sslCertFilePath,
              sslKeyFilePath=platformCfg.sslKeyFilePath)

    VortexFactory.createTcpServer(name=PeekPlatformConfig.componentName,
                                  port=PeekPlatformConfig.config.peekServerVortexTcpPort)


def main():
    setupPlatform()
    from peek_platform import PeekPlatformConfig
    import peek_server

    cfg = PeekPlatformConfig.config

    # Configure sqlalchemy
    setupDbConn(
        metadata=metadata,
        dbEngineArgs=cfg.dbEngineArgs,
        dbConnectString=cfg.dbConnectString,
        alembicDir=os.path.join(os.path.dirname(peek_server.__file__), "alembic")
    )

    # Force model migration
    from peek_server.storage import dbConn
    dbConn.migrate()

    # Import remaining components
    importPackages()

    ###########################################################################
    # BEGIN - INTERIM STORAGE SETUP
    # Force model migration

    from peek_storage._private.storage import setupDbConn as storage_setupDbConn
    from peek_storage import _private as storage_private

    from peek_storage._private.storage.DeclarativeBase import metadata as storage_metadata
    storage_setupDbConn(
        metadata=storage_metadata,
        dbEngineArgs=cfg.dbEngineArgs,
        dbConnectString=cfg.dbConnectString,
        alembicDir=os.path.join(os.path.dirname(storage_private.__file__), "alembic"))

    from peek_storage._private.storage import dbConn as storage_dbConn
    storage_dbConn.migrate()

    # END - INTERIM STORAGE SETUP
    ###########################################################################

    ###########################################################################
    # BEGIN - PATCH CELERY TASKS TO RUN IN PostGreSQL

    # Setup TX Celery
    if cfg.celeryPlPythonEnablePatch:
        from peek_platform.CeleryPatchToPlPython import _DeferredTaskPatch

        _DeferredTaskPatch \
            .setupPostGreSQLConnection(storage_dbConn.ormSessionCreator,
                                       cfg.dbConnectString,
                                       parallelism=cfg.celeryPlPythonWorkerCount)

    else:
        from txcelery.defer import _DeferredTask
        _DeferredTask.startCeleryThreads(
            cfg.celeryConnectionPoolSize,
            cfg.celeryConnectionRecycleTime)

    # END - PATCH CELERY TASKS TO RUN IN PostGreSQL
    ###########################################################################

    reactor.addSystemEventTrigger('before', 'shutdown',
                                  PeekPlatformConfig.pluginLoader.stopOptionalPlugins)
    reactor.addSystemEventTrigger('before', 'shutdown',
                                  PeekPlatformConfig.pluginLoader.stopCorePlugins)

    reactor.addSystemEventTrigger('before', 'shutdown',
                                  PeekPlatformConfig.pluginLoader.unloadOptionalPlugins)
    reactor.addSystemEventTrigger('before', 'shutdown',
                                  PeekPlatformConfig.pluginLoader.unloadCorePlugins)

    # Load all plugins
    d = PeekPlatformConfig.pluginLoader.loadCorePlugins()
    d.addCallback(lambda _: PeekPlatformConfig.pluginLoader.loadOptionalPlugins())
    d.addCallback(lambda _: startListening())
    d.addCallback(lambda _: PeekPlatformConfig.pluginLoader.startCorePlugins())
    d.addCallback(lambda _: PeekPlatformConfig.pluginLoader.startOptionalPlugins())

    def startedSuccessfully(_):
        logger.info('Peek Server is running, version=%s',
                    cfg.platformVersion)
        return _

    d.addErrback(vortexLogFailure, logger, consumeError=True)
    d.addCallback(startedSuccessfully)

    reactor.run()


if __name__ == '__main__':
    main()





class AdminUserAccess:
    """ Admin User Access
    This class stores the details about the user that are required by the
    elements and resources to be able to allow access.
    """

    def __init__(self):
        """
        Constructor
        """
        self.loggedIn = False
        self.username = 'noname'

    def __repr__(self):
        return str(dict(
            loggedIn=self.loggedIn,
            username=self.username,
        ))



from peek_server.backend.auth.AdminAuthRealm import IAuth
from peek_server.backend.navbar.PeekAdmNavbarUserTuple import PeekAdmNavbarUserTuple
from twisted.internet.defer import inlineCallbacks
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope

navbarDataFiltKey = {
    'plugin': 'peek_server',
    'key': "nav.adm.user.data"
}


# HANDLER
class __NavbarHandler:
    def __init__(self):
        self._ep = PayloadEndpoint(navbarDataFiltKey, self._process)

    def shutdown(self):
        self._ep.shutdown()

    @inlineCallbacks
    def _process(self, payloadEnvelope: PayloadEnvelope,
                 httpSession, sendResponse, **kwargs):
        userAccess = IAuth(httpSession).userAccess

        if payloadEnvelope.filt.get('logout'):
            httpSession.expire()
            userAccess.loggedIn = False
            return

        navbarTuple = PeekAdmNavbarUserTuple()
        navbarTuple.username = userAccess.username

        encodedPayload = yield Payload(navbarDataFiltKey, [navbarTuple]) \
            .toEncodedPayloadDefer()

        payloadEnvelope = PayloadEnvelope(navbarDataFiltKey, encodedPayload)
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()

        sendResponse(vortexMsg)


navbarDataHandler = __NavbarHandler()




import logging
import os

from twisted.web._element import renderer, Element
from twisted.web.template import XMLFile
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)


class AdminAuthElement(Element):
    xmlFileName = 'AdminAuthTemplate.xml'
    loader = XMLFile(os.path.join(os.path.dirname(__file__), xmlFileName))

    def __init__(self, failed, failureMsg):
        self._failed = failed
        self._failureMsg = failureMsg

    @renderer
    @deferToThreadWrapWithLogger(logger)
    def loginTitle(self, request, tag):
        return tag(b"Login to Peek Admin")

    @renderer
    @deferToThreadWrapWithLogger(logger)
    def metaDescriptionContent(self, request, tag):
        return tag(content=b"...")

    @renderer
    def errorPanel(self, request, tag):
        if self._failed:
            if self._failureMsg:
                return tag(b"Failed to login : %s" % self._failureMsg.encode('utf-8'))
            return tag(b"Failed to login")
        return b""





from vortex.Tuple import addTupleType, Tuple, TupleField


@addTupleType
class PeekAdmNavbarUserTuple(Tuple):
    __tupleType__ = 'peek_server.PeekAdmNavbarUserTuple'

    username = TupleField('nouser')


from twisted.cred.error import LoginFailed
from twisted.internet.defer import Deferred, inlineCallbacks
from txhttputil.site.AuthCredentials import AuthCredentials


class AdminAuthChecker(AuthCredentials):

    @inlineCallbacks
    def check(self, username, password) -> Deferred:

        from peek_platform import PeekPlatformConfig
        from peek_server.plugin.PeekServerPlatformHook import PeekServerPlatformHook

        coreUserAuthAdmin = PeekServerPlatformHook(PeekPlatformConfig.pluginLoader) \
            .getOtherPluginApi("peek_core_user") \
            .adminAuth

        if username == PeekPlatformConfig.config.adminUser:
            if password != PeekPlatformConfig.config.adminPass:
                raise LoginFailed("Password incorrect for Admin recovery user")
            return

        yield coreUserAuthAdmin.check(username, password)



import logging

from peek_server.backend.auth.AdminAuthChecker import AdminAuthChecker
from peek_server.backend.auth.AdminAuthResource import LoginSucceededResource, \
    LoginResource
from peek_server.backend.auth.AdminUserAccess import AdminUserAccess
from twisted.python.components import registerAdapter
from twisted.web.resource import IResource
from twisted.web.server import Session
from twisted.web.util import DeferredResource
from zope.interface import Interface, Attribute, implementer

logger = logging.getLogger(name="AuthRealm")

__author__ = 'synerty'


class IAuth(Interface):
    value = Attribute("An int value which counts up once per page view.")


@implementer(IAuth)
class Auth:

    def __init__(self, session):
        self.userAccess = AdminUserAccess()
        self.userAccess.loggedIn = False
        self.userAccess.readOnly = True


registerAdapter(Auth, Session, IAuth)


@implementer(IResource)
class AdminAuthRealm:
    isLeaf = False

    def __init__(self, rootResource, authChecker: AdminAuthChecker):
        self._rootResource = rootResource
        self._authChecker = authChecker

    def _authorizedResource(self, request):
        """
        Get the L{IResource} which the given request is authorized to receive.
        If the proper authorization headers are present, the resource will be
        requested from the portal.  If not, an anonymous login attempt will be
        made.
        """

        # The Angular web app does not need authentication
        # if request.uri.endswith(b".js") or request.uri.endswith(b".html"):
        #     return self._rootResource

        auth = IAuth(request.getSession())
        if auth.userAccess.loggedIn:
            request.getSession().touch()
            return self._rootResource

        # TRY form based authentication
        a = request.args
        if request.method == b"POST":
            if b"username" in a and b"password" in a:
                user = a[b'username'][0].decode()
                pass_ = a[b"password"][0].decode()

                def loginCb(_):
                    auth.userAccess = AdminUserAccess()
                    auth.userAccess.username = user
                    auth.userAccess.loggedIn = True
                    return LoginSucceededResource()

                d = self._authChecker.check(user, pass_)
                d.addCallback(loginCb)
                d.addErrback(lambda failure: LoginResource(str(failure.value)))
                return DeferredResource(d)

        # Return login form
        return LoginResource()

    def render(self, request):
        """
        Find the L{IResource} avatar suitable for the given request, if
        possible, and render it.  Otherwise, perhaps render an error page
        requiring authorization or describing an internal server failure.
        """
        return self._authorizedResource(request).render(request)

    def getChildWithDefault(self, path, request):
        """
        Inspect the Authorization HTTP header, and return a deferred which,
        when fired after successful authentication, will return an authorized
        C{Avatar}. On authentication failure, an C{UnauthorizedResource} will
        be returned, essentially halting further dispatch on the wrapped
        resource and all children
        """
        # Don't consume any segments of the request - this class should be
        # transparent!
        request.postpath.insert(0, request.prepath.pop())
        return self._authorizedResource(request)



"""
 * view.common.uiobj.Style.py
 *
 *  Copyright Synerty Pty Ltd 2017
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by
 *  Synerty Pty Ltd
 *
"""
from peek_server.backend.auth.AdminAuthElement import AdminAuthElement
from twisted.python.failure import Failure
from twisted.web._flatten import flattenString
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET
from twisted.web.util import redirectTo


class ElementRenderedCallback():
    def __init__(self, request):
        self.request = request

    def elementRenderedCallback(self, s):
        if self.request.finished or not self.request.channel:
            return

        if isinstance(s, Failure):
            if s.printDetailedTraceback():
                self.request.write(s.printDetailedTraceback())
            s.raiseException()

        else:
            self.request.write(s)

        self.request.finish()


class LoginResource(Resource):
    isLeaf = True

    def __init__(self, failureMsg=None):
        self._failureMsg = failureMsg

    def render_GET(self, request):
        return self._renderLogin(request)

    def render_POST(self, request):
        return self._renderLogin(request, failed=True)

    def _renderLogin(self, request, failed=False):
        request.responseHeaders.setRawHeaders(b"authorization", [b"basic"])

        callback = ElementRenderedCallback(request).elementRenderedCallback

        request.write(b'<!DOCTYPE html>\n')

        d = flattenString(request, AdminAuthElement(failed=failed,
                                                    failureMsg=self._failureMsg))
        d.addBoth(callback)
        return NOT_DONE_YET


class LoginSucceededResource(Resource):
    isLeaf = True

    def render_GET(self, request):
        return self._render(request)

    def render_POST(self, request):
        return self._render(request)

    def _render(self, request):
        return redirectTo(b'/', request)
